import sqlite3
import zlib
from pathlib import Path
from typing import IO, Dict, Union

import numpy as np
import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_mod_mass, internal_without_mods

from .spectral_library import SpectralLibrary

DLIB_COL_NAMES = [
    "MassArray",
    "IntensityArray",
    "MassEncodedLength",
    "IntensityEncodedLength",
    "PrecursorCharge",
    "PeptideModSeq",
    "PeptideSeq",
    "RTInSeconds",
    "PrecursorMz",
]


class DLib(SpectralLibrary):
    """Main to init a DLib obj."""

    def _initialize(self, out: Union[IO, sqlite3.Connection]):
        if isinstance(out, IO):
            raise TypeError("Not supported. Use msp/spectronaut if you want to write a text file.")
        if self.mode == "w":
            DLib._create_database(out)

    def _get_handle(self):
        return sqlite3.connect(self.out_path)

    def _calculate_masked_values(self, fragmentmz: np.ndarray, intensities: np.ndarray):
        """
        Internal function that masks, filters, byte encodes, swaps and compresses fragmentmz \
        and intensities.

        This will produce the data for the following columns in this order:
            - 'MassArray'
            - 'IntensityArray',
            - 'MassEncodedLength',
            - 'IntensityEncodedLength'.
        :param fragmentmz: fragmentmz provided in __init__
        :param intensities: intensities provided in __init__
        :return: 4 lists as described above
        """
        mz_bytes_list = []
        i_bytes_list = []
        mz_lengths = []
        i_lengths = []

        full_mask = self._fragment_filter_passed(fragmentmz, intensities)
        for mz, i, mask in zip(fragmentmz, intensities, np.array(full_mask)):
            # mask to only existing peaks
            sort_index = np.argsort(mz[mask])
            masked_mz_ordered = mz[mask][sort_index]
            masked_i_ordered = i[mask][sort_index]

            # convert arrays to big-endian byte blops
            masked_mz_ordered = np.array(masked_mz_ordered, dtype=np.double)
            masked_i_ordered = np.array(masked_i_ordered * 100, dtype=np.float32)
            masked_mz_ordered.byteswap(inplace=True)
            masked_i_ordered.byteswap(inplace=True)
            bytes_mz = bytes(masked_mz_ordered)
            bytes_i = bytes(masked_i_ordered)
            mz_bytes_list.append(zlib.compress(bytes_mz))
            i_bytes_list.append(zlib.compress(bytes_i))
            mz_lengths.append(len(bytes_mz))
            i_lengths.append(len(bytes_i))
        return mz_bytes_list, i_bytes_list, mz_lengths, i_lengths

    @staticmethod
    def _create_database(conn: sqlite3.Connection):
        """
        Creates the database file with prefab tables entries, peptidetoprotein (p2p) and metadata, according to the \
        dlib specification.

        :param conn: specifies the path of the created database file
        """
        sql_create_entries = """
            CREATE TABLE IF NOT EXISTS entries
            (
                PrecursorMz REAL NOT NULL,
                PrecursorCharge INTEGER NOT NULL,
                PeptideModSeq TEXT NOT NULL,
                PeptideSeq TEXT NOT NULL,
                Copies INTEGER NOT NULL DEFAULT 1,
                RTInSeconds REAL NUT NULL,
                Score REAL NOT NULL DEFAULT 0,
                MassEncodedLength INTEGER NOT NULL,
                MassArray BLOB NOT NULL,
                IntensityEncodedLength INTEGER NOT NULL,
                IntensityArray BLOB NOT NULL,
                CorrelationEncodedLength INTEGER,
                CorrelationArray BLOB,
                RTInSecondsStart REAL,
                RTInSecondsStop REAL,
                MedianChromatogramEncodedLength INTEGER,
                MedianChromatogramArray BLOB,
                SourceFile TEXT NOT NULL DEFAULT 'Oktoberfest'
            )
        """
        sql_create_p2p = """
            CREATE TABLE IF NOT EXISTS peptidetoprotein
            (
                PeptideSeq TEXT NOT NULL,
                isDecoy BOOL DEFAULT FALSE,
                ProteinAccession TEXT NOT NULL DEFAULT 'UNKNOWN'
            )
        """
        sql_create_meta = """
            CREATE TABLE IF NOT EXISTS metadata (Key string not null, Value string not null)
        """
        sql_insert_meta = "INSERT INTO metadata VALUES (?,?)"
        c = conn.cursor()
        c.execute(sql_create_entries)
        c.execute(sql_create_p2p)
        c.execute(sql_create_meta)
        c.execute(sql_insert_meta, ["version", "0.1.14"])
        c.execute(sql_insert_meta, ["staleProteinMapping", "true"])
        conn.commit()

    def _write(self, out: Union[IO, sqlite3.Connection], data: Dict[str, np.ndarray], metadata: pd.DataFrame):
        if isinstance(out, IO):
            raise TypeError("Not supported. Use msp/spectronaut if you want to write a text file.")
        seqs = metadata["SEQUENCE"]
        modseqs = metadata["MODIFIED_SEQUENCE"]
        mass_mod_sequences = internal_to_mod_mass(modseqs)

        p_charges = metadata["PRECURSOR_CHARGE"]
        p_mzs = (metadata["MASS"] + (p_charges * PARTICLE_MASSES["PROTON"])) / p_charges
        # ces = metadata["COLLISION_ENERGY"]

        pr_ids = metadata["PROTEINS"]

        # prepare spectra
        irts = data["irt"][:, 0]  # should create a 1D view of the (n_peptides, 1) shaped array
        f_mzss = data["mz"]
        f_intss = data["intensities"]
        # f_annotss = data["annotation"].astype("S", copy=False)

        masked_values = self._calculate_masked_values(f_mzss, f_intss)

        data_list = [*masked_values, p_charges, mass_mod_sequences, seqs, irts, p_mzs]
        entries = pd.DataFrame(dict(zip(DLIB_COL_NAMES, data_list)))
        p2p = pd.DataFrame({"PeptideSeq": seqs, "ProteinAccession": pr_ids})

        out.execute("BEGIN")

        entries.to_sql(index=False, name="entries", con=out, if_exists="append", method="multi")
        p2p.to_sql(index=False, name="peptidetoprotein", con=out, if_exists="append", method="multi")

        out.commit()
        # conn.close()

        # def prepare_spectrum(self):
        """Converts grpc output and metadata dataframe into dlib format."""
        # precursor_mz: Union[List[float], np.ndarray],
        # precursor_charges: Union[List[int], np.ndarray],
        # modified_sequences: List[str],
        # retention_times: Union[List[float], np.ndarray],
        # fragmentmz: List[np.ndarray],
        # intensities: List[np.ndarray],

        # intensities = self.grpc_output[list(self.grpc_output)[0]]["intensity"]
        # fragment_mz = self.grpc_output[list(self.grpc_output)[0]]["fragmentmz"]
        # annotation = self.grpc_output[list(self.grpc_output)[0]]["annotation"]
        # irt = self.grpc_output[list(self.grpc_output)[1]]
        # retention_times = irt.flatten()
        # modified_sequences = self.spectra_input["MODIFIED_SEQUENCE"]

        # precursor_charges = self.spectra_input["PRECURSOR_CHARGE"]
        # precursor_masses = self.spectra_input["MASS"]
        # precursor_mz = (precursor_masses + (precursor_charges * PARTICLE_MASSES["PROTON"])) / precursor_charges

        # self.create_database(self.out_path)

        # gather all values for the entries table and create pandas DataFrame
        # masked_values = self._calculate_masked_values(fragment_mz, intensities)
        # mass_mod_sequences = internal_to_mod_mass(modified_sequences)
        # sequences = internal_without_mods(modified_sequences)
        # data_list = [*masked_values, precursor_charges, mass_mod_sequences, sequences, retention_times, precursor_mz]
        # self.entries = pd.DataFrame(dict(zip(DLIB_COL_NAMES, data_list)))

        # hardcoded entries that we currently not use.
        # Visit https://bitbucket.org/searleb/encyclopedia/wiki/EncyclopeDIA%20File%20Formats for dlib specs
        # self.entries["Copies"] = 1  # this is hardcorded for now and unused
        # self.entries["Score"] = 0
        # self.entries["CorrelationEncodedLength"] = None
        # self.entries["CorrelationArray"] = None
        # self.entries["RTInSecondsStart"] = None
        # self.entries["RTInSecondsStop"] = None
        # self.entries["MedianChromatogramEncodedLength"] = None
        # self.entries["MedianChromatogramArray"] = None
        # self.entries["SourceFile"] = "Prosit"

        # gather all values for the p2p table and create pandas DataFrame
        # self.p2p = pd.DataFrame({"PeptideSeq": sequences, "isDecoy": False, "ProteinAccession": "unknown"})



from pathlib import Path
from typing import Union

import pandas as pd


def read_file(path: Union[str, Path]) -> pd.DataFrame:
    """
    Read csv file and return df with contents.

    :param path: path to file to read
    :return: df with contents as pd.DataFrame
    """
    df = pd.read_csv(path, sep=",")
    return df


def write_file(df: pd.DataFrame, path: Union[str, Path]):
    """
    Write dataframe to csv file.

    :param df: df with contents as pd.DataFrame
    :param path: path to file to write
    """
    df.to_csv(path, index=False)



import logging
from pathlib import Path
from typing import List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import scipy

# TODO add sparse matrix / anndata support
# TODO add speed benchmarks
# TODO add support for HuggingFace datasets API

Pathlike = Union[Path, str]
Dataset = Union[pd.DataFrame, scipy.sparse.spmatrix]

logger = logging.getLogger(__name__)


def read_file(path: Pathlike) -> pd.DataFrame:
    """
    Read a Parquet file and return a Pandas DataFrame with its contents.

    :param path: Path to the Parquet file to read
    :return: a Pandas DataFrame with the contents of the file
    """
    return pd.read_parquet(path)


def read_partition(path: Pathlike, dataset_name: str) -> pd.DataFrame:
    """
    Read a single table from a partitioned dataset.

    :param path: Root path of the partitioned dataset
    :param dataset_name: Name of the dataset to extract

    :return: a Pandas DataFrame of the specified table from the partitioned dataset
    """
    try:
        dataset = pq.ParquetDataset(path, filters=[("dataset", "=", dataset_name)])
        df = dataset.read().to_pandas().drop("dataset", axis=1)
    except pa.lib.ArrowNotImplementedError as e1:
        if dataset_name.isdigit():
            try:
                logger.warning("Failed to read from partition using string of integer as key, trying with int...")
                dataset = pq.ParquetDataset(path, filters=[("dataset", "=", int(dataset_name))])
                df = dataset.read().to_pandas().drop("dataset", axis=1)
            except pa.lib.ArrowNotImplementedError as e2:
                logger.exception(e2)
        else:
            logger.exception(e1)

    return df


def write_file(data: Dataset, path: Pathlike) -> None:
    """Writes a single DataFrame or matrix to a Parquet file.

    :param data: Data to store
    :param path: Path to write the Parquet file to

    :raises NotImplementedError: if anything else but a Pandas DataFrame is used as the dataset
    """
    if isinstance(data, pd.DataFrame):
        data.to_parquet(path)
    else:
        raise NotImplementedError


def write_partition(datasets: List[Dataset], path: Pathlike, dataset_names: List[str]) -> None:
    """
    Write several datasets to a Parquet dataset as a directory containing subdirectories partinioned by dataset name.

    :param datasets: Datasets to write
    :param path: Root path to write the partitioned dataset to
    :param dataset_names: Names to assign to the datasets for retrieval. Careful: If all of these are strings of ints,
        Parquet will convert them to raw integers!

    :raises NotImplementedError: if anything else but a Pandas DataFrame is used as the dataset
    """
    if all(isinstance(x, pd.DataFrame) for x in datasets):
        df = pd.concat([dataset.assign(dataset=name) for dataset, name in zip(datasets, dataset_names)])
        table = pa.Table.from_pandas(df)
    else:
        raise NotImplementedError

    if isinstance(path, str):
        path = Path(path)
    path.mkdir(exist_ok=True)

    pq.write_to_dataset(
        table,
        root_path=path,
        partition_cols=["dataset"],
        existing_data_behavior="delete_matching",
    )



import glob
import logging
import os
import re
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import xisearch_to_internal

from .search_results import SearchResults

logger = logging.getLogger(__name__)


class Xisearch(SearchResults):
    """Handle search results from xisearch."""

    def read_result(self, tmt_labeled: str = "") -> pd.DataFrame:
        """
        Function to read a csv of CSMs and perform some basic formatting.

        :param tmt_labeled: tmt label as str
        :raises NotImplementedError: if a tmt label is provided
        :return: pd.DataFrame with the formatted data
        """
        if tmt_labeled != "":
            raise NotImplementedError("TMT is not supported for XIsearch")

        logger.info("Reading search results file...")
        columns_to_read = [
            "run_name",
            "scan_number",
            "precursor_mass",
            "precursor_charge",
            "crosslinker_name",
            "decoy_p1",
            "base_sequence_p1",
            "aa_len_p1",
            "link_pos_p1",
            "linked_aa_p1",
            "mods_p1",
            "mod_pos_p1",
            "decoy_p2",
            "base_sequence_p2",
            "aa_len_p2",
            "link_pos_p2",
            "linked_aa_p2",
            "mods_p2",
            "mod_pos_p2",
            "linear",
            "match_score",
        ]

        converters = {"mods_p1": str, "mods_p2": str, "mod_pos_p1": str, "mod_pos_p2": str}

        df = pd.read_csv(self.path, sep="\t", usecols=columns_to_read, converters=converters)
        logger.info("Finished reading search results file.")
        # Standardize column names
        df = Xisearch.filter_xisearch_result(df)
        df = Xisearch.update_columns_for_prosit(df)
        df = Xisearch.filter_valid_prosit_sequences(df)
        return df

    @staticmethod
    def filter_xisearch_result(df: pd.DataFrame) -> pd.DataFrame:
        """
        Remove unsupported modifications and keep only k-k as linked amino acid .

        :param df: df to filter
        :return: filtered df as pd.DataFrame
        """
        df = df[~df["linear"]]
        df = df[df["linked_aa_p1"].notna() & df["linked_aa_p1"].str.contains("K")]
        df = df[df["linked_aa_p2"].notna() & df["linked_aa_p2"].str.contains("K")]
        df = df[~df["mods_p1"].str.contains("dsso-hyd", na=False)]
        df = df[~df["mods_p2"].str.contains("dsso-hyd", na=False)]

        return df

    @staticmethod
    def update_columns_for_prosit(df: pd.DataFrame) -> pd.DataFrame:
        """
        Update columns of df to work with xl-prosit.

        :param df: df to modify
        :return: modified df as pd.DataFrame
        """
        df["decoy"] = df["decoy_p1"] | df["decoy_p2"]
        df["RAW_FILE"] = df["run_name"]
        df["MASS"] = df["precursor_mass"]
        df["PRECURSOR_CHARGE"] = df["precursor_charge"]
        df["CROSSLINKER_TYPE"] = df["crosslinker_name"]
        df["SCORE"] = df["match_score"]
        df["REVERSE"] = df["decoy"]
        df["SCAN_NUMBER"] = df["scan_number"]
        df["SEQUENCE_A"] = df["base_sequence_p1"]
        df["SEQUENCE_B"] = df["base_sequence_p2"]
        df["Modifications_A"] = df["mods_p1"]
        df["Modifications_B"] = df["mods_p2"]
        df["CROSSLINKER_POSITION_A"] = df["link_pos_p1"]
        df["CROSSLINKER_POSITION_B"] = df["link_pos_p2"]
        df["ModificationPositions1"] = df["mod_pos_p1"]
        df["ModificationPositions2"] = df["mod_pos_p2"]
        df["PEPTIDE_LENGTH_A"] = df["aa_len_p1"]
        df["PEPTIDE_LENGTH_B"] = df["aa_len_p2"]
        logger.info("Converting XIsearch peptide sequence to internal format...")

        df["RAW_FILE"] = df["RAW_FILE"].str.replace(".raw", "")

        df["MODIFIED_SEQUENCE_A"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_A"],
                mod=row["Modifications_A"],
                crosslinker_position=row["CROSSLINKER_POSITION_A"],
                mod_positions=row["ModificationPositions1"],
            ),
            axis=1,
            result_type="expand",
        )

        df["MODIFIED_SEQUENCE_B"] = df.apply(
            lambda row: xisearch_to_internal(
                xl=row["CROSSLINKER_TYPE"],
                seq=row["SEQUENCE_B"],
                mod=row["Modifications_B"],
                crosslinker_position=row["CROSSLINKER_POSITION_B"],
                mod_positions=row["ModificationPositions2"],
            ),
            axis=1,
            result_type="expand",
        )

        return df

    @staticmethod
    def filter_valid_prosit_sequences(df: pd.DataFrame) -> pd.DataFrame:
        """
        Filter valid Prosit sequences.

        :param df: df to filter
        :return: df after filtration
        """
        logger.info(f"#sequences before filtering for valid prosit sequences: {len(df.index)}")

        df = df[(df["PEPTIDE_LENGTH_A"] <= 30)]
        df = df[df["PEPTIDE_LENGTH_A"] >= 6]
        df = df[(df["PEPTIDE_LENGTH_B"] <= 30)]
        df = df[df["PEPTIDE_LENGTH_B"] >= 6]
        df = df[(~df["SEQUENCE_A"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        df = df[(~df["SEQUENCE_B"].str.contains(r"B|\*|\.|U|O|X|Z|\(|\)"))]
        df = df[df["PRECURSOR_CHARGE"] <= 6]
        logger.info(f"#sequences after filtering for valid prosit sequences: {len(df.index)}")

        return df



import logging
import warnings
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from xml.etree import ElementTree

import pandas as pd
import pymzml
from pyteomics import mzml
from spectrum_fundamentals.constants import MZML_DATA_COLUMNS

logger = logging.getLogger(__name__)


def check_analyzer(mass_analyzers: Dict[str, str]) -> Dict[str, str]:
    """
    Convert mass analyzer accession ids to internal format.

    :param mass_analyzers: dictionary with instrumentConfigurationRef, analyzer accession
    :return: dictionary with instrumentConfigurationRef, one of "ITMS", "FTMS", "TOF" or "unknown",
        in case the analyzer accession cannot be mapped to one of the three groups.
    """
    for elem in mass_analyzers.keys():
        accession = mass_analyzers[elem]
        if accession in ["MS:1000079", "MS:1000484"]:  # fourier transform ion cyclotron, orbitrap
            mass_analyzers[elem] = "FTMS"
        elif accession in ["MS:1000082", "MS:1000264", "MS:1000078"]:  # quadrupole ion-trap, ion-trap, linear ion-trap
            mass_analyzers[elem] = "ITMS"
        elif accession in ["MS:1000084"]:  # TOF
            mass_analyzers[elem] = "TOF"
        else:
            mass_analyzers[elem] = "unsupported"
    return mass_analyzers


def get_mass_analyzer(file_path: Path) -> Dict[str, str]:
    """
    Retrieve mass analyzer information from mzml file.

    This is using the description of the mzml format to check for specific accessions in the mzml file
    that are not covered by pyteomics or pymzml. The documentation can be found here:
    https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo

    :param file_path: The path to the mzml file to parse
    :return: A dictionary with instrumentConfigurationId, mass analyzer (ITMS, FTMS or TOF) to represent
        the respective mass analyzer category for each MS level that is present in the mzml file, i.e. MS1/MS2/MS3.
    """
    context = ElementTree.iterparse(file_path, events=("start", "end"))
    _, root = next(context)  # Get the root element

    namespace = "{http://psi.hupo.org/ms/mzml}"
    mass_analyzers = {}

    config_id = None
    accession = None
    within_instrument_configuration_list = False
    within_instrument_configuration = False

    for event, element in context:
        if event == "start" and element.tag == f"{namespace}instrumentConfigurationList":
            within_instrument_configuration_list = True
        elif event == "end" and element.tag == f"{namespace}instrumentConfigurationList":
            within_instrument_configuration_list = False
            break
        elif (
            within_instrument_configuration_list
            and event == "start"
            and element.tag == f"{namespace}instrumentConfiguration"
        ):
            within_instrument_configuration = True
            config_id = element.attrib.get("id")
        elif (
            within_instrument_configuration
            and element.tag == f"{namespace}analyzer"
            and element.attrib.get("order") == "2"
            and event == "start"
        ):
            accession = None  # Reset accession
        elif (
            within_instrument_configuration
            and element.tag == f"{namespace}cvParam"
            and element.get("accession") is not None
        ):
            if accession is None:
                accession = element.attrib.get("accession")
        elif (
            within_instrument_configuration and event == "end" and element.tag == f"{namespace}instrumentConfiguration"
        ):
            within_instrument_configuration = False
            if config_id is not None and accession is not None:
                mass_analyzers[config_id] = accession

    return check_analyzer(mass_analyzers)


class MSRaw:
    """Main to read mzml file and generate dataframe containing intensities and m/z values."""

    def __init__(self, path: Optional[Union[str, Path]] = None, output_path: Optional[Union[str, Path]] = None):
        """
        Initialize a MSRaw object.

        :param path: path to mzml file
        :param output_path: path to save the output file
        """
        if isinstance(path, str):
            path = Path(path)
        if isinstance(output_path, str):
            output_path = Path(output_path)
        self.path = path
        self.output_path = output_path

    @staticmethod
    def read_mzml(
        source: Union[str, Path, List[Union[str, Path]]],
        ext: str = "mzml",
        package: str = "pyteomics",
        search_type: str = "Maxquant",
        scanidx: Optional[List] = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Reads mzml and generates a dataframe containing intensities and m/z values.

        :param source: a directory containing mzml files, a list of files or a single file
        :param ext: file extension for searching a specified directory
        :param package: package for parsing the mzml file. Can eiter be "pymzml" or "pyteomics"
        :param scanidx: optional list of scan numbers to extract. if not specified, all scans will be extracted
        :param search_type: type of the search (Maxquant, Mascot, Msfragger)
        :param args: additional positional arguments
        :param kwargs: additional keyword arguments
        :raises AssertionError: if package has an unexpected type
        :return: pd.DataFrame with intensities and m/z values
        """
        file_list = MSRaw.get_file_list(source, ext)

        if package == "pymzml":
            data = MSRaw._read_mzml_pymzml(file_list, scanidx, *args, **kwargs)
        elif package == "pyteomics":
            data = MSRaw._read_mzml_pyteomics(file_list, *args, **kwargs)
        else:
            raise AssertionError("Choose either 'pymzml' or 'pyteomics'")

        data["SCAN_NUMBER"] = pd.to_numeric(data["SCAN_NUMBER"])
        return data

    @staticmethod
    def _read_mzml_pymzml(file_list: List[Path], scanidx: Optional[List] = None, *args, **kwargs) -> pd.DataFrame:
        data_dict = {}
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ImportWarning)
            for file_path in file_list:
                logger.info(f"Reading mzML file: {file_path}")
                data_iter = pymzml.run.Reader(file_path, args=args, kwargs=kwargs)
                file_name = file_path.stem
                mass_analyzer = get_mass_analyzer(file_path)
                namespace = "{http://psi.hupo.org/ms/mzml}"
                instrument_name = data_iter.info["referenceable_param_group_list_element"][0][0].get("name")

                if scanidx is None:
                    spectra = data_iter
                else:
                    # this does not work if some spectra are filtered out, e.g. mzML files with only MS2 spectra, see:
                    # https://github.com/pymzml/pymzML/blob/a883ff0e61fd97465b0a74667233ff594238e335/pymzml/file_classes
                    # /standardMzml.py#L81-L84
                    spectra = (data_iter[idx] for idx in scanidx)

                for spec in spectra:
                    if spec.ms_level != 2:
                        continue  # filter out ms1 spectra if there are any
                    key = f"{file_name}_{spec.ID}"
                    scan = spec.get_element_by_path(["scanList", "scan"])[0]
                    instrument_configuration_ref = scan.get("instrumentConfigurationRef", "")
                    activation = spec.get_element_by_path(["precursorList", "precursor", "activation"])[0]
                    fragmentation = "unknown"
                    collision_energy = 0.0
                    for cv_param in activation:
                        name = cv_param.get("name")
                        if name == "collision energy":
                            collision_energy = float(cv_param.get("value"))
                            continue
                        if "beam-type" in name:
                            fragmentation = "HCD"
                        elif "collision-induced dissociation" in name:
                            fragmentation = "CID"
                        else:
                            fragmentation = name
                    scan_window = scan.find(f".//{namespace}scanWindow")
                    scan_lower_limit = float(
                        scan_window.find(f'./{namespace}cvParam[@accession="MS:1000501"]').get("value")
                    )
                    scan_upper_limit = float(
                        scan_window.find(f'./{namespace}cvParam[@accession="MS:1000500"]').get("value")
                    )
                    mz_range = f"{scan_lower_limit}-{scan_upper_limit}"
                    data_dict[key] = [
                        file_name,
                        spec.ID,
                        spec.i,
                        spec.mz,
                        mz_range,
                        spec.scan_time_in_minutes(),
                        mass_analyzer.get(instrument_configuration_ref, "unknown"),
                        fragmentation,
                        collision_energy,
                        instrument_name,
                    ]
                data_iter.close()
        data = pd.DataFrame.from_dict(data_dict, orient="index", columns=MZML_DATA_COLUMNS)
        return data

    @staticmethod
    def _read_mzml_pyteomics(file_list: List[Path], *args, **kwargs) -> pd.DataFrame:
        data_dict = {}
        for file_path in file_list:
            mass_analyzer = get_mass_analyzer(file_path)
            logger.info(f"Reading mzML file: {file_path}")
            data_iter = mzml.read(str(file_path), *args, **kwargs)
            file_name = file_path.stem
            try:
                instrument_params = data_iter.get_by_id("commonInstrumentParams")
            except KeyError:
                instrument_params = data_iter.get_by_id("CommonInstrumentParams")
            instrument_name = list(instrument_params.keys())[1]
            for spec in data_iter:
                if spec["ms level"] != 2:
                    continue  # filter out ms1 spectra if there are any
                spec_id = spec["id"].split("scan=")[-1]
                scan = spec["scanList"]["scan"][0]
                instrument_configuration_ref = scan.get("instrumentConfigurationRef", "")
                activation = spec["precursorList"]["precursor"][0]["activation"]
                fragmentation = "unknown"
                collision_energy = 0.0
                for key, value in activation.items():
                    if key == "collision energy":
                        collision_energy = value
                    elif "beam-type" in key:
                        fragmentation = "HCD"
                    elif "collision-induced dissociation" in key:
                        fragmentation = "CID"
                    else:
                        fragmentation = key
                scan_lower_limit = scan["scanWindowList"]["scanWindow"][0]["scan window lower limit"]
                scan_upper_limit = scan["scanWindowList"]["scanWindow"][0]["scan window upper limit"]
                mz_range = f"{scan_lower_limit}-{scan_upper_limit}"
                rt = spec["scanList"]["scan"][0]["scan start time"]
                key = f"{file_name}_{spec_id}"
                data_dict[key] = [
                    file_name,
                    spec_id,
                    spec["intensity array"],
                    spec["m/z array"],
                    mz_range,
                    rt,
                    mass_analyzer.get(instrument_configuration_ref, "unknown"),
                    fragmentation,
                    collision_energy,
                    instrument_name,
                ]
            data_iter.close()
        data = pd.DataFrame.from_dict(data_dict, orient="index", columns=MZML_DATA_COLUMNS)
        return data

    @staticmethod
    def get_file_list(source: Union[str, Path, List[Union[str, Path]]], ext: str = "mzml") -> List[Path]:
        """
        Get list of files from source.

        :param source: a directory containing mzml files, a list of files or a single file
        :param ext: file extension for searching a specified directory
        :raises FileNotFoundError: if one of the files given by source does not exist
        :raises TypeError: if source is not given as a str, Path or list object
        :return: list of files
        """
        file_list = []
        if isinstance(source, str):
            source = Path(source)
        if isinstance(source, Path):
            if source.is_file():
                file_list = [source]
            elif source.is_dir():
                file_list = list(source.glob("*[mM][zZ][mM][lL]"))
            else:
                raise FileNotFoundError(f"{source} does not exist.")
        elif isinstance(source, list):
            for elem in source:
                if isinstance(elem, str):
                    elem = Path(elem)
                if elem.is_file():
                    file_list.append(elem)
                else:
                    raise FileNotFoundError(f"{elem} does not exist,")
        else:
            raise TypeError("source can only be a single str or Path or a list of files.")
        return file_list



"""Top-level package for spectrum_io."""

__author__ = "Mario Picciani"
__email__ = "mario.picciani@tum.de"
__version__ = "0.5.0"

import logging
import logging.handlers
import sys
import time

from . import d, file, raw
from .search_result import MaxQuant
from .spectral_library import DLib, Spectronaut

CONSOLE_LOG_LEVEL = logging.INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
if len(logger.handlers) == 0:
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s::%(funcName)s %(message)s")
    converter = time.gmtime
    # add console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(CONSOLE_LOG_LEVEL)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # add error handler
    error_handler = logging.StreamHandler()
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(formatter)
    logger.addHandler(error_handler)
else:
    logger.info("Logger already initizalized. Resuming normal operation.")



import os
import re
from itertools import chain, cycle
from sqlite3 import Connection
from typing import IO, Dict, Tuple, Union

import numpy as np
import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_spectronaut, internal_without_mods

from .spectral_library import SpectralLibrary


class Spectronaut(SpectralLibrary):
    """Main to initialze a Spectronaut obj."""

    # Check spectronaut folder for output format.

    @staticmethod
    def _assemble_fragment_string(f_int: float, f_mz: float, f_annot: bytes):
        m = re.match(r"([by])(\d+)\+(\d)(?:-(\w+))?", f_annot.decode())
        if m is None:
            raise ValueError(f"Malformed annotation string encountered: {f_annot.decode()}")
        return (
            f"{f_int:.4f},{f_mz:.8f},{m.group(2)},{m.group(1)},{m.group(3)},{m.group(4) if m.group(4) else 'noloss'}\n"
        )

    def _write(self, out: Union[IO, Connection], data: Dict[str, np.ndarray], metadata: pd.DataFrame):
        # prepare metadata
        if isinstance(out, Connection):
            raise TypeError("Not supported. Use DLib if you want to write a database file.")
        seqs = metadata["SEQUENCE"]
        modseqs = internal_to_spectronaut(metadata["MODIFIED_SEQUENCE"].apply(lambda x: "_" + x + "_"))
        p_charges = metadata["PRECURSOR_CHARGE"]
        p_mzs = (metadata["MASS"] + (p_charges * PARTICLE_MASSES["PROTON"])) / p_charges
        ces = metadata["COLLISION_ENERGY"]
        pr_ids = metadata["PROTEINS"]

        # prepare spectra
        irts = data["irt"][:, 0]  # should create a 1D view of the (n_peptides, 1) shaped array
        f_mzss = data["mz"]
        f_intss = data["intensities"]
        f_annotss = data["annotation"].astype("S", copy=False)

        vec_assemble = np.vectorize(Spectronaut._assemble_fragment_string)

        for f_ints, f_mzs, modseq, seq, p_charge, p_mz, irt, ce, pr_id, f_annots in zip(
            f_intss, f_mzss, modseqs, seqs, p_charges, p_mzs, irts, ces, pr_ids, f_annotss
        ):
            cond = self._fragment_filter_passed(f_mzs, f_ints)
            line_start = [f"{modseq},{seq},{seq},{p_charge},{p_mz:.8f},{irt:.2f},{ce},{pr_id},"]
            fragment_list = vec_assemble(f_ints[cond], f_mzs[cond], f_annots[cond])
            out.writelines(chain.from_iterable(zip(cycle(line_start), fragment_list)))

    def _initialize(self, out: Union[IO, Connection]):
        if isinstance(out, Connection):
            raise TypeError("Not supported. Use DLib if you want to write a database file.")
        if self.mode == "w":
            out.write(
                "ModifiedPeptide,LabeledPeptide,StrippedPeptide,PrecursorCharge,PrecursorMz,iRT,CollisionEnergy,ProteinIds,"
                "RelativeFragmentIntensity,FragmentMz,FragmentNumber,FragmentType,FragmentCharge,FragmentLossType\n"
            )



from abc import abstractmethod
from multiprocessing import Queue
from multiprocessing.managers import ValueProxy
from pathlib import Path
from sqlite3 import Connection
from typing import IO, Dict, Optional, Union

import numpy as np
import pandas as pd


class SpectralLibrary:
    """Main to initialze a SpectralLibrary obj."""

    def __init__(
        self,
        output_path: Union[str, Path],
        mode: str = "w",
        min_intensity_threshold: float = 5e-4,
        chunksize: Optional[int] = None,
    ):
        """
        Initialize a SpectralLibrary obj.

        :param output_path: path to output file including file name
        :param mode: Whether to append ('a') to or overwrite ('w') an extisting file
            at the provided output path (if it is present).
        :param min_intensity_threshold: optional filter for low intensity peaks
        :param chunksize: optional chunksize for dlib
        """
        if isinstance(output_path, str):
            output_path = Path(output_path)
        self.out_path = output_path
        self.mode = mode
        self.min_intensity_threshold = min_intensity_threshold
        self.chunksize = chunksize

    def load(self):
        """Load predictions from hdf5 file."""

    def write(self, *args, **kwargs):
        """
        Write content to the output file.

        :param args: Positional arguments to be passed to the internal _write method.
        :param kwargs: Keyword arguments to be passed to the internal _write method.
        """
        with self._get_handle() as out:
            self._initialize(out)
            self._write(out, *args, **kwargs)

    def _get_handle(self):
        return open(self.out_path, self.mode)

    def async_write(self, queue: Queue, progress: ValueProxy):
        """
        Asynchronously write content to the output file from a queue.

        :param queue: A queue from which content will be retrieved for writing.
        :param progress: An integer value representing the progress of the writing process.
        """
        with self._get_handle() as out:
            self._initialize(out)
            while True:
                content = queue.get()
                if content is None:
                    break
                self._write(out, *content)
                progress.value += 1

    def _fragment_filter_passed(
        self, f_mz: Union[np.ndarray, float], f_int: Union[np.ndarray, float]
    ) -> Union[np.ndarray, bool]:
        """
        Return if a fragment passes a common filter.

        This function returns a boolean to determine if a fragment can exist (as defined by its mz != -1)
        and if its (predicted) intensity is at least the minimal threshold given during instantiation of
        the writer class.

        :param f_mz: The fragment's mass to charge ratio
        :param f_int: The fragment's (predicted) intensity

        :return: boolean (array) that determines if the fragment passes the filter
        """
        return (f_mz != -1) & (f_int >= self.min_intensity_threshold)

    @abstractmethod
    def _write(self, out: Union[IO, Connection], data: Dict[str, np.ndarray], metadata: pd.DataFrame):
        """
        Internal writer function.

        This function takes a batch of data and corresponding metadata and writes it to the provided
        output file handle. The file handle logic is provided using the public write or async_write
        functions.

        :param out: file handle accepting the data to be written to disk
        :param data: Dictionary containing TODO keys and corresponding values as numpy array
        :param metadata: a dataframe that contains the columns TODO
        """
        pass

    @abstractmethod
    def _initialize(self, out: Union[IO, Connection]):
        pass



import math
from typing import Any, Callable, List, Optional

# import pyximport; pyximport.install()
# from mgf_filter.cython.mat import ceil


class Peak:
    """Container class for a single fragment peak."""

    def __init__(self, mz: float, intensity: float, delta_function: Callable, meta: Optional[List[Any]] = None):
        """
        Contructor for a single fragment peak.

        A fragment peak is contructed from an mz and intensity value, as well as a delta function providing a
        function to calculate the mass tolerance window and an additional metadata list # TODO provide details
        :param mz: mass to charge ratio of peak
        :param intensity: relative intensity of the peak
        :param delta_function: callable to calculate mass tolerance window
        :param meta: metadata list
        """
        self.mz = mz
        self.intensity = intensity
        self.left = 0.0
        self.delta_function = delta_function
        self.counts = 1
        self.meta = meta if meta is not None else []
        self.update()

    def __str__(self):
        """Overrides the default str method to provide information about mz, intensity, left border, and delta."""
        return (
            "mz: "
            + str(self.mz)
            + "\n"
            + "intensity: "
            + str(self.intensity)
            + "\n"
            + "left: "
            + str(self.left)
            + "\n"
            + "delta: "
            + str(self.delta)
        )

    def update(self):
        """Updates delta and calculate left border."""
        # left is needed for key()
        # function will be overwritten
        self.delta = self.delta_function(self.mz)
        self.left = self.mz - self.delta
        self.ceiled_key = math.ceil(self.left)

    def key(self):
        """Provides access to the ceiled key, which is ceil rounded left border."""
        return self.ceiled_key



import logging
import re
from abc import abstractmethod
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from spectrum_io.file import csv

logger = logging.getLogger(__name__)


def filter_valid_prosit_sequences(df: pd.DataFrame) -> pd.DataFrame:
    """
    Filter valid Prosit sequences.

    :param df: df to filter
    :return: df after filtering out unsupported peptides
    """
    logger.info(f"#sequences before filtering for valid prosit sequences: {len(df.index)}")
    # retain only peptides that fall within [7, 30] length supported by Prosit
    df = df[(df["PEPTIDE_LENGTH"] <= 30) & (df["PEPTIDE_LENGTH"] >= 7)]
    # remove unsupported mods to exclude
    unsupported_mods = [r"Acetyl \(Protein N\-term\)", "ac", r"\[[0-9]+\]", r"\+"]
    exclude_mods_pattern = re.compile("|".join(unsupported_mods))
    df = df[~df["MODIFIED_SEQUENCE"].str.contains(exclude_mods_pattern)]
    # remove non-canonical aas
    df = df[(~df["SEQUENCE"].str.contains("U|O"))]
    # remove precursor charges greater than 6
    df = df[df["PRECURSOR_CHARGE"] <= 6]
    logger.info(f"#sequences after filtering for valid prosit sequences: {len(df.index)}")

    return df


class SearchResults:
    """Handle search results from different software."""

    orig_res: pd.DataFrame
    fake_msms: pd.DataFrame

    def __init__(self, path: Union[str, Path]):
        """
        Init Searchresults object.

        :param path: path to file
        """
        if isinstance(path, str):
            path = Path(path)
        self.path = path

    @abstractmethod
    def read_result(self, tmt_labeled: str):
        """Read result.

        :param tmt_labeled: tmt label as str

        """
        raise NotImplementedError

    def generate_internal(self, tmt_labeled: str, out_path: Optional[Union[str, Path]] = None) -> pd.DataFrame:
        """
        Generate df and save to out_path if provided.

        :param out_path: path to output
        :param tmt_labeled: tmt label as str
        :return: path to output file
        """
        if out_path is None:
            # convert and return
            return self.read_result(tmt_labeled)

        if isinstance(out_path, str):
            out_path = Path(out_path)

        if out_path.is_file():
            # only read converted and return
            logger.info(f"Found search results in internal format at {out_path}, skipping conversion")
            return csv.read_file(out_path)

        # convert, save and return
        df = self.read_result(tmt_labeled)
        csv.write_file(df, out_path)
        return df

    def read_internal(self) -> pd.DataFrame:
        """
        Read file from path.

        :return: dataframe after reading the file
        """
        return csv.read_file(self.path)



import logging
import sqlite3
from pathlib import Path
from typing import Optional, Union

import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import internal_without_mods

from .search_results import SearchResults, filter_valid_prosit_sequences

logger = logging.getLogger(__name__)


class Mascot(SearchResults):
    """Handle search results from Mascot."""

    def read_result(self, tmt_labeled: str) -> pd.DataFrame:
        """
        Function to read a mascot msf file and perform some basic formatting.

        :param tmt_labeled: tmt label as str
        :return: pd.DataFrame with the formatted data
        """
        logger.info("Reading mascot msf file")
        connection = sqlite3.connect(self.path)
        # cursor = connection.cursor()
        # cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        df = pd.read_sql("SELECT * FROM MSnSpectrumInfo", connection)[
            ["SpectrumID", "SpectrumFileName", "RetentionTime", "Mass", "Charge"]
        ]
        df_id_map = pd.read_sql("SELECT * FROM TargetPsmsMSnSpectrumInfo", connection)[
            ["MSnSpectrumInfoSpectrumID", "TargetPsmsPeptideID"]
        ]
        df = df.merge(df_id_map, left_on="SpectrumID", right_on="MSnSpectrumInfoSpectrumID")
        df_target_psms = pd.read_sql("SELECT * FROM TargetPsms", connection)[
            ["PeptideID", "Sequence", "ModifiedSequence", "Modifications", "XCorr"]
        ]
        df = df.merge(df_target_psms, left_on="TargetPsmsPeptideID", right_on="PeptideID")
        df_modif = pd.read_sql("SELECT * FROM TargetPsmsFoundModifications", connection)[
            ["TargetPsmsPeptideID", "FoundModificationsModificationID", "Position"]
        ]
        df_modif_mass = pd.read_sql("SELECT * FROM FoundModifications", connection)[
            ["ModificationID", "DeltaMonoisotopicMass"]
        ]
        df_modif = df_modif.merge(df_modif_mass, left_on="FoundModificationsModificationID", right_on="ModificationID")
        df = df.merge(df_modif, on="TargetPsmsPeptideID")

        logger.info("Finished reading mascot msf file.")

        df.rename(
            columns={
                "SpectrumID": "SCAN NUMBER",
                "ModifiedSequence": "MODIFIED SEQUENCE",
                "Charge": "PRECURSOR CHARGE",
                "XCorr": "SCORE",
                "SpectrumFileName": "RAW FILE",
            },
            inplace=True,
        )

        # Standardize column names
        df.columns = df.columns.str.upper()
        df.columns = df.columns.str.replace(" ", "_")
        # TODO reverse
        df["REVERSE"] = df["SEQUENCE"].str.contains("Reverse")
        logger.info("Converting MSFragger  peptide sequence to internal format")
        df["RAW_FILE"] = df["RAW_FILE"].str.replace(".raw", "")
        df["MODIFICATIONS"] = (
            (df["POSITION"].astype(int) - 1).astype(str) + "$" + df["DELTAMONOISOTOPICMASS"].astype(str)
        )
        df = df.groupby("SCAN_NUMBER", as_index=False).apply(lambda x: x.sort_values("POSITION"))
        df = df.groupby(
            ["SCAN_NUMBER", "PRECURSOR_CHARGE", "SCORE", "RAW_FILE", "SEQUENCE", "REVERSE"],
            as_index=False,
        ).agg({"MODIFICATIONS": "|".join})
        mod_masses_reverse = {round(float(v), 3): k for k, v in c.MOD_MASSES.items()}

        sequences = []
        for _, row in df.iterrows():
            modifications = row["MODIFICATIONS"].split("|")
            if len(modifications) == 0:
                sequences.append(row["SEQUENCE"])
            else:
                sequence = row["SEQUENCE"]
                skip = 0
                for mod in modifications:
                    pos, mass = mod.split("$")
                    sequence = (
                        sequence[: int(pos) + 1 + skip]
                        + mod_masses_reverse[round(float(mass), 3)]
                        + sequence[int(pos) + 1 + skip :]
                    )
                    skip = skip + len(mod_masses_reverse[round(float(mass), 3)])
                sequences.append(sequence)

        df["MODIFIED_SEQUENCE"] = sequences

        df["SEQUENCE"] = internal_without_mods(df["MODIFIED_SEQUENCE"])
        df["PEPTIDE_LENGTH"] = df["SEQUENCE"].apply(lambda x: len(x))

        return filter_valid_prosit_sequences(df)



import logging
import subprocess
from pathlib import Path
from sys import platform
from typing import Any, List, Optional, Tuple, Union

from .msraw import MSRaw

logger = logging.getLogger(__name__)


def _type_check(var: Any, varname: str, types: Union[type, Tuple[type, ...]]):
    if isinstance(var, types):
        return
    if isinstance(types, type):
        possible_types_str = types.__name__
    else:
        type_names = [t.__name__ for t in types]
        possible_types_str = "{} or {}".format(", ".join(type_names[:-1]), type_names[-1])
    raise TypeError(f"{varname} must be of type {possible_types_str}")


def _assemble_arg_list(
    input_path: Path, output_path: Path, ms_level: List[int], gzip: bool, thermo_exe: Path
) -> List[Union[str, Path]]:
    exec_arg_list: List[Union[str, Path]] = [
        thermo_exe,
        f"--msLevel={','.join([str(l) for l in ms_level])}",
        "-i",
        input_path.resolve(),
        "-b",
        output_path,
    ]
    if gzip:
        exec_arg_list.append("-g")
    if "linux" in platform or platform == "darwin":
        exec_arg_list.insert(0, "mono")

    return exec_arg_list


class ThermoRaw(MSRaw):
    """Main to convert a ThermoRaw file into mzml file."""

    @staticmethod
    def convert_raw_mzml(
        input_path: Union[Path, str],
        gzip: bool = False,
        ms_level: Union[int, List[int]] = 2,
        output_path: Optional[Union[Path, str]] = None,
        thermo_exe: Union[Path, str] = "ThermoRawFileParser.exe",
    ) -> Path:
        """Converts a ThermoRaw file to mzML.

        Use https://github.com/compomics/ThermoRawFileParser for conversion.

        :param input_path: file path of the Thermo Rawfile
        :param gzip: whether to gzip the file
        :param ms_level: level of MS, can be a single integer (1, 2, 3) or any combination of that provided as a list
        :param output_path: file path of the mzML path
        :param thermo_exe: path to the executable of ThermoRawFileParser. Default: ThermoRawFileParser.exe
        :raises subprocess.CalledProcessError: if the subprocess for conversion failed
        :raises ValueError: if ms_level(s) provided are other than 1, 2 or 3.
        :return: path to converted file as string
        """
        _type_check(input_path, "input_path", (Path, str))
        input_path = Path(input_path)
        if output_path is None:
            output_path = input_path.with_suffix(".mzML")
        _type_check(output_path, "output_path", (Path, str))
        output_path = Path(output_path)

        _type_check(thermo_exe, "thermo_exe", (Path, str))
        thermo_exe = Path(thermo_exe)

        _type_check(ms_level, "ms_level", (int, list))
        if isinstance(ms_level, int):
            ms_level = [ms_level]
        for level in ms_level:
            _type_check(level, "all ms_levels in list", int)
            if not 1 <= level <= 3:
                raise ValueError(f"Value of all ms_levels must be within [1,3]. Got {level}")

        if output_path.is_file():
            logger.info(f"Found converted file at {output_path}, skipping conversion")
            return output_path

        exec_arg_list = _assemble_arg_list(input_path, output_path, ms_level, gzip, thermo_exe)

        logger.info(
            f"Converting thermo rawfile to mzml with the command: {' '.join([str(arg) for arg in exec_arg_list])}"
        )

        try:
            subprocess.run(exec_arg_list, shell=False, check=True)
        except subprocess.CalledProcessError:
            if output_path.is_file():
                output_path.unlink()
            raise  # reraise only after removing a corrupted file

        return output_path


if __name__ == "__main__":
    from sys import argv

    if len(argv) == 2:
        converter = ThermoRaw().convert_raw_mzml(argv[1], ms_level=[1, 2])
    else:
        print("Please specify a rawfile")



import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import alphatims
import alphatims.bruker
import alphatims.utils
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from .masterSpectrum import MasterSpectrum

logger = logging.getLogger(__name__)


def binning(mzs: List[float], intensities: List[int], ignore_charges: bool) -> Tuple[List[float], List[float]]:
    """
    Perform binning on the input MasterSpectrum.

    This function loads a MasterSpectrum from the provided input, performs binning, and exports the results
    to a temporary text file in the given rescoring path. It then reads the temporary file as a DataFrame,
    modifies the DataFrame by adding SCAN_NUMBER and dropping specified columns before returning it.

    :param intensities: Input data used to perform binning.
    :param mzs: Path where the temporary file will be exported.
    :param ignore_charges: indicating whether charges should be ignored during binning.
    :return: Tuple containing the sorted list of fragment mzs and associated intensities
    """
    ms = MasterSpectrum()
    ms.load_from_tims(intensities, mzs, ignore_charges)

    mzs_out = [mp.mz for key in ms.spectrum[0].keys() for mp in ms.spectrum[0][key]]
    intensities_out = [mp.intensity for key in ms.spectrum[0].keys() for mp in ms.spectrum[0][key]]

    return mzs_out, intensities_out


def aggregate_timstof(raw_spectra: pd.DataFrame) -> pd.DataFrame:
    """
    Combine spectra from the provided pd.DataFrame and perform binning on chunks.

    This function splits the input pd.DataFrame into chunks, performs binning on each chunk of data,
    merges the binning results with the original data, processes the combined spectra, and returns
    the combined and processed spectra as a pd.DataFrame.

    :param raw_spectra: pd.DataFrame containing spectra information.
    :return: pd.DataFrame containing combined and processed spectra.
    """
    for i, (combined_intensities, combined_mzs) in tqdm(
        enumerate(zip(raw_spectra["INTENSITIES"], raw_spectra["MZ"])),
        total=len(raw_spectra),
        desc="Aggregating spectra",
    ):
        mz, intensity = binning(combined_mzs, combined_intensities, True)
        raw_spectra.at[i, "INTENSITIES"] = intensity
        raw_spectra.at[i, "MZ"] = mz

    return raw_spectra


def read_timstof(hdf_file: Path, scan_to_precursor_map: pd.DataFrame) -> pd.DataFrame:
    """
    Read selected spectra from a given timstof hdf file.

    This function queries a given hdf file for spectra that are provided within a scan to precursor map.
    #TODO elaborate

    :param hdf_file: Path to hdf file containing spectra
    :param scan_to_precursor_map: Dataframe containing metadata to select spectra

    :return: Dataframe containing the relevant spectra read from the hdf file
    """
    # preparation of filter
    df_frame_group = (
        scan_to_precursor_map[["FRAME", "PRECURSOR"]]
        .drop_duplicates()
        .groupby("FRAME", as_index=False)
        .agg(
            {
                "PRECURSOR": tuple,
            }
        )
        .groupby("PRECURSOR", as_index=False)
        .agg({"FRAME": tuple})
    )

    # load filtered stuff
    data = alphatims.bruker.TimsTOF(str(hdf_file), slice_as_dataframe=False)

    raw_idx = []
    for frames, precursors in zip(df_frame_group["FRAME"], df_frame_group["PRECURSOR"]):
        raw_idx.extend(data[frames, :, precursors])

    # read spectra
    df = data.as_dataframe(
        raw_idx,
        raw_indices=False,
        frame_indices=True,
        scan_indices=True,
        quad_indices=False,
        tof_indices=False,
        precursor_indices=True,
        rt_values=True,
        rt_values_min=False,
        mobility_values=True,
        quad_mz_values=False,
        push_indices=False,
        mz_values=True,
        intensity_values=True,
        corrected_intensity_values=False,
        raw_indices_sorted=False,
    )
    df.columns = ["FRAME", "SCAN", "PRECURSOR", "RETENTION_TIME", "INV_ION_MOBILITY", "MZ", "INTENSITIES"]

    # converting RETENTION TIME from seconds to minutes
    df["RETENTION_TIME"] = df["RETENTION_TIME"].div(60)

    # aggregation
    df_combined_grouped = (
        df.merge(
            scan_to_precursor_map[
                ["SCAN_NUM_BEGIN", "SCAN_NUM_END", "PRECURSOR", "FRAME", "COLLISION_ENERGY"]
            ].drop_duplicates()
        )
        .query("SCAN_NUM_BEGIN <= SCAN <= SCAN_NUM_END")  # can probably be skipped
        .groupby(["PRECURSOR", "FRAME"], as_index=False)  # aggregate fragments per precursor in FRAME
        .agg(
            {
                "INTENSITIES": list,
                "MZ": list,
                "RETENTION_TIME": "first",
                "COLLISION_ENERGY": "first",
                "INV_ION_MOBILITY": "first",
            }
        )
        .merge(scan_to_precursor_map.reset_index())
        .groupby("SCAN_NUMBER", as_index=False)  # aggregate PRECURSORS for same SCAN_NUMBER
        .agg(
            COLLISION_ENERGY=("COLLISION_ENERGY", "median"),
            INTENSITIES=("INTENSITIES", lambda x: [item for sublist in x for item in sublist]),
            MZ=("MZ", lambda x: [item for sublist in x for item in sublist]),
            RETENTION_TIME=("RETENTION_TIME", "median"),
            median_INV_ION_MOBILITY=("INV_ION_MOBILITY", "median"),
        )
    )

    return df_combined_grouped


def convert_d_hdf(
    input_path: Union[Path, str],
    output_path: Union[Path, str],
):
    """
    Convert a bruker d folder to hdf format.

    # TODO long description
    :param input_path: Path to the d folder to be converted
    :param output_path: Path to the desired output location of the converted hdf file
    """
    if isinstance(output_path, str):
        output_path = Path(input_path)
    if output_path.is_file():
        logger.info(f"Found converted file at {output_path}, skipping conversion")
        return
    logger.info("Converting bruker d to hdf using alphatims...")
    data = alphatims.bruker.TimsTOF(str(input_path))
    data.save_as_hdf(directory=str(output_path.parent), file_name=str(output_path.name))


def read_and_aggregate_timstof(source: Path, tims_meta_file: Path) -> pd.DataFrame:
    """
    Read raw spectra from timstof hdf spectra file and aggregate to MS2 spectra.

    :param source: Path to the hdf file
    :param tims_meta_file: Path to metadata mapping scan numbers to precursors / frames
    :return: Dataframe containing the MS2 spectra
    """
    scan_to_precursor_map = pd.read_csv(tims_meta_file)
    raw_spectra = read_timstof(source, scan_to_precursor_map)
    df_combined = aggregate_timstof(raw_spectra)
    df_combined["RAW_FILE"] = source.stem
    df_combined["MASS_ANALYZER"] = "TOF"
    df_combined["FRAGMENTATION"] = "HCD"
    df_combined["INSTRUMENT_TYPES"] = "TIMSTOF"

    return df_combined
